@@ -1,7 +1,5 @@
-import sys
-import re
+import sys, re, copy
 import numpy as np
-import copy
 
 import cctk
 from cctk import Molecule
@@ -10,7 +8,7 @@ from cctk.helper_functions import align_matrices
 
 class Ensemble:
     """
-    Class that represents a group of molecules. They do not all need to have the same atoms or bonds.
+    Class representing a collection of molecules. They do not all need to have the same atoms or bonds.
 
     Ensembles are composed of molecules and properties. Molecules are ``Molecule`` objects, whereas properties are ``dict`` objects containing calculation-specific information.
 
@@ -18,7 +16,7 @@ class Ensemble:
 
     - ``ensemble[molecule]`` or ``ensemble[0]`` will return new ``Ensemble`` objects with only the specified molecules.
         Lists or slices can also be used: so ``ensemble[0:10:2]`` or ``ensemble[[molecule1, molecule2, molecule3]]`` will also return new ``Ensemble`` objects.
-    - Individual properties can be read through tuple indexing: ``ensemble[0,"energy"]`` will return the energy of the first molecule, 
+    - Individual properties can be read through tuple indexing: ``ensemble[0,"energy"]`` will return the energy of the first molecule,
         while ``ensemble[:,"energy"]`` will return a list of all the energies.
     - To access ``Molecule`` objects, use ``ensemble.molecule``: ``ensemble.molecule[0]`` will return the first object, whereas ``ensemble.molecule[1:3]`` will return a list.
     - ``ensemble.items()`` will return a list of (molecule, property) pairs.
@@ -26,10 +24,8 @@ class Ensemble:
 
     Attributes:
         name (str): name, for identification
-        _items (dict):
-            keys: ``Molecule`` objects
-            values: dictionaries containing properties from each molecule, variable. should always be one layer deep.
-            molecules: special object that accesses the keys
+        _items (dict): keys: ``Molecule`` objects; values: dictionaries containing properties from each molecule, variable. should always be one layer deep.
+        molecules (``MoleculeIndexer``): special object that accesses the keys
     """
 
     def __init__(self, name=None):
@@ -152,7 +148,6 @@ class Ensemble:
                     result.append(None)
         if len(ensemble) == 1:
             if result[0] is None:
-#                raise ValueError(f"ensemble did not contain property '{prop}' for key '{idx}'")
                 return None
             return result[0]
         else:
@@ -163,7 +158,8 @@ class Ensemble:
                     break
             if found_something:
                 return result
-            raise ValueError(f"ensemble did not contain property '{prop}' for any of keys '{idx}'")
+            else:
+                return None
 
     def get_properties_dict(self, idx):
         """
@@ -242,10 +238,33 @@ class Ensemble:
             assert isinstance(num, int), "num must be integer"
             return list(self.values())[num]
 
-    def sort_by(self, property):
-        pass
+    def sort_by(self, property_name, ascending=True):
+        """
+        Sorts the ensemble by the specified property.
+        Throws an error if the property is missing for any entries.
+        Consistent, sort-compatible property values are assumed and not checked.
 
-    def add_molecule(self, molecule, properties={}, copy=False):
+        Args:
+            property_name (str): the name of the property to sort on (must be a string or number)
+            ascending (bool): whether the property should increase or decrease in value
+        Returns:
+            new Ensemble (current ensemble is not modified)
+        """
+        property_list = self[:,property_name]
+        if property_list is None:
+            raise ValueError(f"property '{property_name}' not found in ensemble")
+        property_list = np.asarray(property_list)
+        n_missing_entries = np.count_nonzero(property_list==None)
+        if n_missing_entries > 0:
+            error = "---sorting error---\n"
+            error += str(property_list)
+            raise ValueError(f"{error}\nproperty '{property_name}' has {n_missing_entries} missing entries and cannot be sorted")
+        new_indices = np.argsort(property_list)
+        if not ascending:
+            new_indices = np.flip(new_indices)
+        return self[[new_indices]]
+
+    def add_molecule(self, molecule, properties=None, copy=False):
         """
         Adds a molecule to the ensemble.
 
@@ -256,10 +275,17 @@ class Ensemble:
         """
         if not isinstance(molecule, Molecule):
             raise TypeError("molecule is not a Molecule - so it can't be added!")
-        assert isinstance(properties, dict), f"properties must be a dict and not type {type(properties)}"
 
         if copy:
             molecule = copy.deepcopy(molecule)
+
+        if properties is None:
+            #### empty dicts all point to the same memory address by default, so need to prevent that behavior by initializing non-empty dict
+            properties = {"placeholder": 1}
+            del properties["placeholder"]
+
+        assert isinstance(properties, dict), f"properties must be a dict and not type {type(properties)}"
+
         self._items[molecule] = properties
 
     def _check_molecule_number(self, number):
@@ -294,11 +320,27 @@ class Ensemble:
 
         return new_ensemble
 
+    def lowest_molecules(self, property_name, num=1):
+        """
+        Retrieves the molecules with the lowest values of the specified property.
+
+        Args:
+            property_name (str): the name of the property to sort on
+            num (int): how many molecules to return
+        Returns:
+            lowest ``Molecule`` (if num==1)
+            ``list`` of ``Molecule`` (otherwise)
+        """
+        assert isinstance(num, (int, np.integer)), f"num must be an integer, got {type(num)}"
+        assert num > 0, f"num must be > 0, got {num}"
+        sorted_ensemble = self.sort_by(property_name)
+        if num > 1:
+            return sorted_ensemble.molecules[0:num]
+        return sorted_ensemble.molecules[0]
 
 class ConformationalEnsemble(Ensemble):
     """
-    Class that represents a group of conformers. All members must have the same atom types in the same order.
-
+    Class that representing a group of conformers. All members must have the same atom types in the same order.
     """
 
     def __str__(self):
@@ -311,7 +353,7 @@ class ConformationalEnsemble(Ensemble):
         else:
             return f"ConformationalEnsemble ({len(self._items)} molecules, {n_atoms} atoms)"
 
-    def add_molecule(self, molecule, properties={}, copy=False):
+    def add_molecule(self, molecule, properties=None, copy=False, checks=True):
         """
         Checks that the molecule contains the same atom types in the same order as existing molecules, and that the molecule has the same charge/multiplicity.
         """
@@ -326,7 +368,7 @@ class ConformationalEnsemble(Ensemble):
             if molecule.multiplicity != initial_mol.multiplicity:
                 raise ValueError("wrong spin multiplicity for this ensemble")
 
-            if not np.array_equal(molecule.atomic_numbers, initial_mol.atomic_numbers):
+            if checks and not np.array_equal(molecule.atomic_numbers, initial_mol.atomic_numbers):
                 raise ValueError("wrong atom types for this ensemble")
 
             #### only save one copy to save space
@@ -431,7 +473,7 @@ class ConformationalEnsemble(Ensemble):
         """
         Aligns every geometry in this ensemble and then creates a new ensemble that contains only the non-redundant conformers.
         If energies are available, the lowest energy conformer will be kept for every redundancy.
-        The current ensemble will not be modified.
+        The current ensemble will not be modified.  The resulting ensemble will be sorted by energy (if available).
 
         Args:
             RMSD_cutoff (float): remove conformers that are more similar than this threshold
@@ -522,8 +564,4 @@ class ConformationalEnsemble(Ensemble):
 
         return output
 
-    def get_lowest_energy(self, num=10):
-        return self.molecules[np.argsort(self.energies)][0:10]
 
-    def get_within_cutoff(self, cutoff=5):
-        return self.molecules[self.energies <= (np.min(self.energies) + 5)]
