@@ -4,14 +4,16 @@ import numpy as np
 from enum import Enum
 
 from cctk import File, Ensemble, Molecule, ConformationalEnsemble, OneIndexedArray
-from cctk.helper_functions import get_symbol, compute_distance_between, compute_angle_between, compute_dihedral_between, get_number
+from cctk.helper_functions import get_symbol, compute_distance_between, \
+                                  compute_angle_between, compute_dihedral_between, \
+                                  get_number, get_corrected_free_energy
 
 import cctk.parse_gaussian as parse
 
 
-class JobType(Enum):
+class GaussianJobType(Enum):
     """
-    Class to contain allowed Gaussian job types. Not an exhaustive list, but should be fairly comprehensive.
+    Class representing allowed Gaussian job types. Not an exhaustive list, but should be fairly comprehensive.
 
     The value should be the Gaussian keyword, to permit automatic assignment.
 
@@ -19,17 +21,44 @@ class JobType(Enum):
     """
 
     SP = "sp"
+    """
+    Single point energy calculation.
+    """
+
     OPT = "opt"
+    """
+    Geometry optimization.
+    """
+
     FREQ = "freq"
+    """
+    Hessian calculation.
+    """
+
     IRC = "irc"
+    """
+    Intrinsic reaction coordinate calculation.
+    """
+
     NMR = "nmr"
+    """
+    NMR shielding prediction.
+    """
+
     POP = "pop"
+    """
+    Population analysis.
+    """
+
     FORCE = "force"
+    """
+    Gradient calculation.
+    """
 
 #### This static variable tells what properties are expected from each JobType.
 EXPECTED_PROPERTIES = {
     "sp": ["energy", "scf_iterations",],
-    "opt": ["rms_displacement", "rms_force", ],
+    "opt": ["rms_displacement", "rms_force",],
     "freq": ["gibbs_free_energy", "enthalpy", "frequencies",],
     "nmr": ["isotropic_shielding",],
     "pop": [],
@@ -39,32 +68,34 @@ EXPECTED_PROPERTIES = {
 
 class GaussianFile(File):
     """
-    Class for Gaussian files. Composes ``ConformationalEnsemble``.
+    Class representing Gaussian input/output files.
 
     Attributes:
         ensemble (Ensemble): ``ConformationalEnsemble`` instance
-        job_types (list): list of `job_type` instances
+        job_types (list): list of `job_type`` instances
         route_card (str): optional, route card of .gjf file
         link0 (dict): optional, dictionary of Link 0 commands (e.g. {"mem": "32GB", "nprocshared": 16})
         footer (str): optional, footer of .gjf file
-        success (int): number of successful terminations (should be 1 for an opt, 2 for opt and then freq, 1 for a single point energy, etc)
+        successful_terminations (int): number of successful terminations (should be 1 for an opt, 2 for opt and then freq, 1 for a single point energy, etc)
+        elapsed_time (float): total time for job in seconds
         title (str): optional, title of .gjf file
     """
 
     def __init__(
-        self, job_types, route_card=None, link0=None, footer=None, title="title", success=0
+        self, job_types, route_card=None, link0=None, footer=None, title="title", success=0, elapsed_time=0.0
     ):
         """
         Create new GaussianFile object.
 
-	Args:
+        Args:
             job_types (list): list of ``job_type`` instances
             route_card (str): optional, route card of ``.gjf`` file
             link0 (dict): optional, Link 0 commands of ``.gjf`` file
             footer (str): optional, footer of ``.gjf`` file
             title (str): optional, title of ``.gjf`` file
             success (int): num successful terminations
-	"""
+            elapsed_time (float): total time for job in seconds
+        """
 
         if route_card and not isinstance(route_card, str):
             raise TypeError("route card needs to be a string")
@@ -81,8 +112,11 @@ class GaussianFile(File):
         if success and not isinstance(success, int):
             raise TypeError("success needs to be an integer")
 
+        if not isinstance(elapsed_time, float) or elapsed_time < 0.0:
+            raise TypeError(f"elapsed_time invalid: {elapsed_time}")
+
         if job_types is not None:
-            if not all(isinstance(job, JobType) for job in job_types):
+            if not all(isinstance(job, GaussianJobType) for job in job_types):
                 raise TypeError(f"invalid job type {job}")
 
         self.ensemble = ConformationalEnsemble()
@@ -91,7 +125,8 @@ class GaussianFile(File):
         self.footer = footer
         self.title = title
         self.job_types = job_types
-        self.success = success
+        self.successful_terminations = success
+        self.elapsed_time = elapsed_time
 
     def __str__(self):
         return f"GaussianFile (title=\"{str(self.title)}\", {len(self.ensemble)} entries in Ensemble)"
@@ -186,7 +221,7 @@ class GaussianFile(File):
         """
         Returns the imaginary frequencies, rounded to the nearest integer.
         """
-        if (JobType.FREQ in self.job_types) and (self.ensemble[-1:,"frequencies"] is not None):
+        if (GaussianJobType.FREQ in self.job_types) and (self.ensemble[-1:,"frequencies"] is not None):
             freqs = self.ensemble[-1:,"frequencies"]
             if not isinstance(freqs, list) or len(freqs) == 0:
                 return list()
@@ -196,7 +231,7 @@ class GaussianFile(File):
             return list()
 
     @classmethod
-    def read_file(cls, filename, return_lines=False):
+    def read_file(cls, filename, return_lines=False, extended_opt_info=False):
         """
         Reads a Gaussian``.out`` or ``.gjf`` file and populates the attributes accordingly.
         Only footers from ``opt=modredundant`` can be read automatically --  ``genecep`` custom basis sets, &c must be specified manually.
@@ -208,6 +243,8 @@ class GaussianFile(File):
         Args:
             filename (str): path to the out file
             return_lines (Bool): whether the lines of the file should be returned
+            extended_opt_info (Bool): if full parameters about each opt step should be collected
+                (by default, only ``rms_displacement`` and ``rms_force`` are collected)
         Returns:
             ``GaussianFile`` object (or list of ``GaussianFile`` objects for Link1 files)
             (optional) the lines of the file (or list of lines of file for Link1 files)
@@ -215,22 +252,38 @@ class GaussianFile(File):
         if re.search("gjf$", filename) or re.search("com$", filename):
             return cls._read_gjf_file(filename, return_lines)
 
-        link1_lines = parse.split_link1(super().read_file(filename))
+        link1_lines = parse.split_link1(filename)
         files = []
         for link1idx, lines in enumerate(link1_lines):
             #### automatically assign job types based on header
-            header = parse.search_for_block(lines, "#p", "----")
+            header = lines.search_for_block("#p", "----", format_line=lambda x: x.lstrip(), join="")
             if header is None:
                 raise ValueError("can't find route card! (perhaps '#p' wasn't employed?)")
             job_types = cls._assign_job_types(header)
 
             link0 = parse.extract_link0(lines)
 
+            title = ""
+            title_block = lines.search_for_block("l101.exe", "Symbolic Z-matrix", join="\n")
+            for line in title_block.split("\n")[1:]:
+                if not re.search("-----", line):
+                    title += line
+
             #### extract parameters
             success = 0
+            elapsed_time = 0.0  # in seconds
             for line in lines:
                 if line.strip().startswith("Normal termination"):
                     success += 1
+                elif line.startswith(" Elapsed time:"):
+                    # Elapsed time:       0 days  0 hours  0 minutes  0.6 seconds.
+                    fields = line.split()
+                    assert len(fields) == 10, f"unexpected number of fields on elapsed time line:\n{line}"
+                    days = float(fields[2])
+                    hours = float(fields[4])
+                    minutes = float(fields[6])
+                    seconds = float(fields[8])
+                    elapsed_time += days * 86400 + hours * 3600 + minutes * 60 + seconds
 
             (geometries, atom_list, energies, scf_iterations,) = parse.read_geometries_and_energies(lines)
             atomic_numbers = []
@@ -243,67 +296,99 @@ class GaussianFile(File):
 
             footer = None
             if re.search("modredundant", str(header)):
-                footer = parse.search_for_block(lines, "^ The following ModRedundant input section", "^ $", count=1, join="\n")
-                footer = "\n".join(list(footer.split("\n"))[1:])  # get rid of the first line
-                footer = "\n".join([" ".join(list(filter(None, line.split(" ")))) for line in footer.split("\n")])
+                footer = lines.search_for_block("^ The following ModRedundant input section", "^ $", count=1, join="\n")
+                if footer is not None:
+                    footer = "\n".join(list(footer.split("\n"))[1:])  # get rid of the first line
+                    footer = "\n".join([" ".join(list(filter(None, line.split(" ")))) for line in footer.split("\n")])
 
             bonds = parse.read_bonds(lines)
-            charge = parse.find_parameter(lines, "Multiplicity", expected_length=4, which_field=1, split_on="=")[0]
-            multip = parse.find_parameter(lines, "Multiplicity", expected_length=4, which_field=3, split_on="=")[0]
+            charge = lines.find_parameter("Multiplicity", expected_length=4, which_field=1, split_on="=")[0]
+            multip = lines.find_parameter("Multiplicity", expected_length=4, which_field=3, split_on="=")[0]
 
-            f = GaussianFile(job_types=job_types, route_card=header, link0=link0, footer=footer, success=success)
+            f = GaussianFile(job_types=job_types, route_card=header, link0=link0, footer=footer, success=success, elapsed_time=elapsed_time, title=title)
 
             molecules = [None] * len(geometries)
             properties = [{} for _ in range(len(geometries))]
             for idx, geom in enumerate(geometries):
                 molecules[idx] = Molecule(atomic_numbers, geom, charge=charge, multiplicity=multip, bonds=bonds)
-                properties[idx]["energy"] = energies[idx]
-                properties[idx]["scf_iterations"] = scf_iterations[idx]
+                if idx < len(energies):
+                    properties[idx]["energy"] = energies[idx]
+                if idx < len(scf_iterations):
+                    properties[idx]["scf_iterations"] = scf_iterations[idx]
                 properties[idx]["link1_idx"] = link1idx
                 properties[idx]["filename"] = filename
+                properties[idx]["iteration"] = idx
 
             #### now for some job-type specific attributes
-            if JobType.OPT in job_types:
-                rms_forces = parse.find_parameter(lines, "RMS\s+Force", expected_length=5, which_field=2)
-                rms_displacements = parse.find_parameter(lines, "RMS\s+Displacement", expected_length=5, which_field=2)
+            if GaussianJobType.OPT in job_types:
+                rms_forces = lines.find_parameter("RMS\s+Force", expected_length=5, which_field=2)
+                rms_displacements = lines.find_parameter("RMS\s+Displacement", expected_length=5, which_field=2)
+
+                if extended_opt_info:
+                    max_forces = lines.find_parameter("Maximum Force", expected_length=5, which_field=2)
+                    max_displacements = lines.find_parameter("Maximum Displacement", expected_length=5, which_field=2)
+                    max_gradients = lines.find_parameter("Cartesian Forces:", expected_length=6, which_field=3)
+                    rms_gradients = lines.find_parameter("Cartesian Forces:", expected_length=6, which_field=5)
+                    max_int_forces = lines.find_parameter("Internal  Forces:", expected_length=6, which_field=3)
+                    rms_int_forces = lines.find_parameter("Internal  Forces:", expected_length=6, which_field=5)
+                    delta_energy = lines.find_parameter("Predicted change in Energy", expected_length=4, which_field=3, cast_to_float=False)
 
                 for idx, force in enumerate(rms_forces):
                     properties[idx]["rms_force"] = force
                     properties[idx]["rms_displacement"] = rms_displacements[idx]
 
-            if JobType.FREQ in job_types:
-                enthalpies = parse.find_parameter(lines, "thermal Enthalpies", expected_length=7, which_field=6)
+                    if extended_opt_info:
+                        properties[idx]["max_force"] = max_forces[idx]
+                        properties[idx]["max_displacement"] = max_displacements[idx]
+                        properties[idx]["max_gradient"] = max_gradients[idx]
+                        properties[idx]["rms_gradient"] = rms_gradients[idx]
+                        properties[idx]["max_internal_force"] = max_int_forces[idx]
+                        properties[idx]["rms_internal_force"] = rms_int_forces[idx]
+                        change_in_energy = re.sub(r"Energy=", "", delta_energy[idx])
+                        properties[idx]["predicted_change_in_energy"] = float(change_in_energy.replace('D', 'E'))
+
+            if GaussianJobType.FREQ in job_types:
+                enthalpies = lines.find_parameter("thermal Enthalpies", expected_length=7, which_field=6)
                 if len(enthalpies) == 1:
                     properties[-1]["enthalpy"] = enthalpies[0]
                 elif len(enthalpies) > 1:
-                    raise ValueError("too many enthalpies found!")
+                    raise ValueError(f"unexpected # of enthalpies found!\nenthalpies = {enthalpies}")
 
-                gibbs_vals = parse.find_parameter(lines, "thermal Free Energies", expected_length=8, which_field=7)
+                gibbs_vals = lines.find_parameter("thermal Free Energies", expected_length=8, which_field=7)
                 if len(gibbs_vals) == 1:
                     properties[-1]["gibbs_free_energy"] = gibbs_vals[0]
                 elif len(gibbs_vals) > 1:
-                    raise ValueError("too many gibbs free energies found!")
+                    raise ValueError(f"unexpected # gibbs free energies found!\ngibbs free energies = {gibbs_vals}")
 
                 frequencies = []
                 try:
-                    frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=2)
-                    frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=3)
-                    frequencies += parse.find_parameter(lines, "Frequencies", expected_length=5, which_field=4)
+                    frequencies += lines.find_parameter("Frequencies", expected_length=5, which_field=2)
+                    frequencies += lines.find_parameter("Frequencies", expected_length=5, which_field=3)
+                    frequencies += lines.find_parameter("Frequencies", expected_length=5, which_field=4)
                     properties[-1]["frequencies"] = sorted(frequencies)
                 except:
                     raise ValueError("error finding frequencies")
 
-            if JobType.NMR in job_types:
+                #  Temperature   298.150 Kelvin.  Pressure   1.00000 Atm.
+                temperature = lines.find_parameter("Temperature", expected_length=6, which_field=1)
+                if len(temperature) == 1:
+                    properties[-1]["temperature"] = temperature[0]
+                    corrected_free_energy = get_corrected_free_energy(gibbs_vals[0], frequencies,
+                                                                      frequency_cutoff=100.0, temperature=temperature[0])
+                    properties[-1]["quasiharmonic_gibbs_free_energy"] = float(corrected_free_energy)
+
+
+            if GaussianJobType.NMR in job_types:
                 assert len(molecules) == 1, "NMR jobs should not be combined with optimizations!"
                 nmr_shifts = parse.read_nmr_shifts(lines, molecules[0].num_atoms())
                 properties[0]["isotropic_shielding"] = nmr_shifts.view(OneIndexedArray)
 
-            if JobType.FORCE in job_types:
+            if GaussianJobType.FORCE in job_types:
                 assert len(molecules) == 1, "force jobs should not be combined with optimizations!"
                 forces = parse.read_forces(lines)
                 properties[0]["forces"] = forces
 
-            if JobType.POP in job_types:
+            if GaussianJobType.POP in job_types:
                 if re.search("hirshfeld", f.route_card):
                     charges, spins = parse.read_hirshfeld_charges(lines)
                     properties[-1]["hirshfeld_charges"] = charges
@@ -442,22 +527,22 @@ class GaussianFile(File):
     @classmethod
     def _assign_job_types(cls, header):
         """
-        Assigns ``JobType`` objects from route card. ``Job.Type.SP`` is assigned by default.
+        Assigns ``GaussianJobType`` objects from route card. ``GaussianJobType.SP`` is assigned by default.
 
-        For instance, "#p opt freq=noraman" would give an output of ``[JobType.SP, JobType.OPT, JobType.FREQ]``.
+        For instance, "#p opt freq=noraman" would give an output of ``[GaussianJobType.SP, GaussianJobType.OPT, GaussianJobType.FREQ]``.
 
         Args:
             header (str): Gaussian route card
 
         Returns:
-            list of ``JobType`` objects
+            list of ``GaussianJobType`` objects
         """
         job_types = []
-        for name, member in JobType.__members__.items():
+        for name, member in GaussianJobType.__members__.items():
             if re.search(f" {member.value}", str(header), re.IGNORECASE):
                 job_types.append(member)
-        if JobType.SP not in job_types:
-            job_types.append(JobType.SP)
+        if GaussianJobType.SP not in job_types:
+            job_types.append(GaussianJobType.SP)
         return job_types
 
     def check_has_properties(self):
@@ -466,11 +551,15 @@ class GaussianFile(File):
 
         This only checks the last molecule in ``self.ensemble``, for now.
         """
-        return True
-#        for job_type in self.job_types:
-#            for prop in EXPECTED_PROPERTIES[job_type.value]:
-#                if not self.ensemble.has_property(-1, prop):
-#                    raise ValueError(f"expected property {prop} for job type {job_type}, but it's not there!")
+        if self.successful_terminations > 0:
+            if self.successful_terminations == 1 and ((GaussianJobType.OPT in self.job_types) and (GaussianJobType.FREQ in self.job_types)):
+                return # opt freq jobs should have two terminations
+            for job_type in self.job_types:
+                for prop in EXPECTED_PROPERTIES[job_type.value]:
+                    if not self.ensemble.has_property(-1, prop):
+                        raise ValueError(f"expected property {prop} for job type {job_type}, but it's not there!")
+        else:
+            return
 
     @classmethod
     def write_ensemble_to_file(cls, filename, ensemble, route_card, link0={"mem": "32GB", "nprocshared": 16}, footer=None, title="title", print_symbol=False):
