@@ -10,7 +10,7 @@ from cctk.helper_functions import get_symbol, get_number
 
 class MOL2File(File):
     """
-    Generic class for all ``.mol2`` files.
+    Class representing SYBYL ``.mol2`` files.
 
     Attributes:
         name (str): name of file
@@ -36,16 +36,43 @@ class MOL2File(File):
 
         file = MOL2File(name=name)
 
-        (geometries, symbols, atom_types, bonds, conformers) = cls._read_mol2(filename, **kwargs)
-        atomic_numbers = np.array([get_number(z) for z in symbols], dtype=np.int8)
+        (geometries, all_clean_symbols, all_symbols, all_bonds, conformers) = cls._read_mol2(filename, **kwargs)
+        assert len(all_bonds) == len(geometries)
+        for bonds in all_bonds:
+            assert isinstance(bonds, nx.Graph)
+            assert len(bonds) == len(geometries[0])
 
-        if conformers == True:
+        if conformers:
+            # convert atom types to atomic numbers
+            atomic_numbers = []
+            for atom_type in all_symbols[0]:
+                assert isinstance(atom_type,str), f"unexpected atom_type type: {type(atom_type)} / {atom_type}"
+                fields = atom_type.split(".")
+                symbol = fields[0]
+                symbol = re.sub("[^A-Za-z]","",symbol)
+                atomic_number = get_number(symbol)
+                atomic_numbers.append(atomic_number)
+            atomic_numbers = np.asarray(atomic_numbers, dtype=np.int8)
+
+            # create ensemble
             file.ensemble = ConformationalEnsemble()
+            for geometry in geometries:
+                molecule = Molecule(atomic_numbers, geometry, bonds=all_bonds[0].edges, checks=False)
+                file.ensemble.add_molecule(molecule, checks=False)
         else:
             file.ensemble = Ensemble()
-
-        for geom in geometries:
-            file.ensemble.add_molecule(Molecule(atomic_numbers, geom, bonds=bonds.edges))
+            for this_symbols,geometry in zip(all_symbols,geometries):
+                atomic_numbers=[]
+                for atom_type in this_symbols:
+                    assert isinstance(atom_type,str), f"unexpected atom_type type: {type(atom_type)} / {atom_type}"
+                    fields = atom_type.split(".")
+                    symbol = fields[0]
+                    symbol = re.sub("[^A-Za-z]","",symbol)
+                    atomic_number = get_number(symbol)
+                    atomic_numbers.append(atomic_number)
+                atomic_numbers = np.asarray(atomic_numbers, dtype=np.int8)
+                molecule = Molecule(atomic_numbers, geometry, bonds=bonds.edges)
+                file.ensemble.add_molecule(molecule)
 
         return file
 
@@ -67,12 +94,6 @@ class MOL2File(File):
                                                 latter option increases performance,
                                                 particularly for large files.
 
-            save_memory_for_conformers (bool): if True, the first dimension (geometry number) will be
-                                            dropped from the symbols and bonds to prevent
-                                            the storage of redundant information.  Thus,
-                                            symbols will be a one-dimensional ``np.ndarray`` of
-                                            ``str`` and bonds will be a single ``nx.Graph``.
-
             print_status_messages (bool): if True, update the progerss of the parsing operation to stdout.
 
         Returns:
@@ -117,6 +138,9 @@ class MOL2File(File):
                 in_bond_block = False
                 i += 1
                 line = lines[i]
+                if contains_conformers == True and len(all_symbols) > 0:
+                    this_symbols = all_symbols[0]
+                    this_clean_symbols = all_clean_symbols[0]
             elif line.startswith("@<TRIPOS>BOND"):
                 # update status
                 in_geometry_block = False
@@ -130,12 +154,15 @@ class MOL2File(File):
                 # initialize connectivity graph
                 if len(this_geometry) == 0:
                     raise ValueError("got to bond table without a geometry")
-                this_bonds = nx.Graph()
-                this_bonds.add_nodes_from(range(1, len(this_geometry) + 1))
+                if contains_conformers == True and len(all_bonds) > 0:
+                    this_bonds = all_bonds[0]
+                else:
+                    this_bonds = nx.Graph()
+                    this_bonds.add_nodes_from(range(1, len(this_geometry) + 1))
 
             # parse geometry if appropriate
             if in_geometry_block:
-                fields = re.split(" +", line.strip())
+                fields = line.split()
                 if len(fields) < 6:
                     print("Error parsing file:")
                     print("Line = '%s'" % line.strip())
@@ -143,13 +170,14 @@ class MOL2File(File):
                     break
                 x, y, z = float(fields[2]), float(fields[3]), float(fields[4])
                 this_geometry.append([x, y, z])
-                symbol = fields[5]
-                clean_symbol = fields[1]
-                this_symbols.append(symbol)
-                this_clean_symbols.append(clean_symbol)
+                if contains_conformers != True or len(all_symbols)==0:
+                    symbol = fields[5]
+                    clean_symbol = fields[1]
+                    this_symbols.append(symbol)
+                    this_clean_symbols.append(clean_symbol)
             elif in_bond_block:
-                fields = re.split(" +", line.strip())
-                if len(fields) == 4:
+                fields = line.split()
+                if len(fields) == 4 and (len(all_bonds)==0 or contains_conformers != True):
                     # parse bonds, checking that the bonds are increasing
                     try:
                         this_bond_number = int(fields[0])
@@ -158,7 +186,10 @@ class MOL2File(File):
                         n_atoms = len(this_geometry)
                         if not 1 <= atom1 <= n_atoms or not 1 <= atom2 <= n_atoms:
                             raise ValueError(f"atom number out of range: {line}")
-                        bond_order = int(fields[3])
+                        if fields[3] == "ar":
+                            bond_order = 1
+                        else:
+                            bond_order = int(fields[3])
                         if bond_order <= 0:
                             raise ValueError(f"zero or negative bond order: {line}")
                         if this_bond_number != bond_number + 1:
@@ -213,19 +244,13 @@ class MOL2File(File):
         else:
             raise ValueError("contains_conformers must be 'check' or boolean")
 
-        # if requested, just store one copy of all_symbols and all_bonds
-        if save_memory_for_conformers and contains_conformers:
-            all_symbols = all_symbols[0]
-            all_clean_symbols = all_clean_symbols[0]
-            all_bonds = all_bonds[0]
-
         # return result
         n_geometries = len(all_geometries)
         if print_status_messages:
             if n_geometries > 1:
                 if contains_conformers:
                     n_atoms = len(all_geometries[0])
-                    n_bonds = all_bonds.number_of_edges()
+                    n_bonds = all_bonds[0].number_of_edges()
                     if print_status_messages:
                         print(f"read {n_geometries} conformers ({n_atoms} atoms and {n_bonds} bonds).")
                 else:
@@ -247,15 +272,11 @@ class MOL2File(File):
                         print(f"read {n_geometries} unrelated geometries ({min_n_atoms}-{max_n_atoms} atoms and {min_n_bonds}-{max_n_bonds}) bonds).")
             else:
                 n_atoms = len(all_geometries)
-                n_bonds = all_bonds.number_of_edges()
+                n_bonds = all_bonds[0].number_of_edges()
                 if print_status_messages:
                     print(f"read one geometry ({n_atoms} atoms and {n_bonds} bonds).")
 
-        #### sometimes these labels switch? so gotta check
-        if len(all_clean_symbols[0]) < len(all_symbols[0]):
-            return (all_geometries, all_clean_symbols, all_symbols, all_bonds, contains_conformers)
-        else:
-            return (all_geometries, all_symbols, all_clean_symbols, all_bonds, contains_conformers)
+        return (all_geometries, all_clean_symbols, all_symbols, all_bonds, contains_conformers)
 
     def get_molecule(self, num=None):
         """
