@@ -1,8 +1,8 @@
-import sys, re, math
+import sys, re, math, copy, scipy
 import numpy as np
 import networkx as nx
-import scipy
 
+import cctk
 from cctk import OneIndexedArray
 from cctk.helper_functions import (
     get_symbol,
@@ -14,11 +14,12 @@ from cctk.helper_functions import (
     compute_unit_vector,
     get_covalent_radius,
     get_isotopic_distribution,
+    compute_chirality,
 )
 
 class Molecule:
     """
-    Class that represents a single molecule, abstractly.
+    Class representing a single molecular geometry.
 
     In contrast to typical Python behavior, ``atomic_numbers`` and ``geometry`` are indexed from one, to simplify interfacing with computational chemistry programs.
     This has been done by defining a custom wrapper for ``numpy.ndarray`` called ``cctk.OneIndexedArray``.
@@ -29,76 +30,84 @@ class Molecule:
         name (str): for identification, optional
         atomic_numbers (cctk.OneIndexedArray, dtype=np.int8): list of atomic numbers
         geometry (cctk.OneIndexedArray): list of 3-tuples of xyz coordinates - same ordering as ``atomic_numbers``
-        bonds (nx.Graph): Graph object containing connectivity information (1-indexed)
+        bonds (nx.Graph or list of tuples): connectivity graph or list of 2-tuples, with each element representing the 1-indexed atom number of a bonded pair
         charge (int): the charge of the molecule
         multiplicity (int): the spin state of the molecule (1 corresponds to singlet, 2 to doublet, 3 to triplet, etc. -- so a multiplicity of 1 is equivalent to S=0)
+        checks (bool): whether to check that the constructor parameters are valid
     """
 
-    def __init__(self, atomic_numbers, geometry, name=None, bonds=None, charge=0, multiplicity=1):
+    def __init__(self, atomic_numbers, geometry, name=None, bonds=None, charge=0, multiplicity=1, checks=True):
         """
         Create new Molecule object, and assign connectivity if needed.
 
         ``bonds`` must be a list of edges (i.e. an n x 2 ``numpy`` array).
         """
-        if len(atomic_numbers) != len(geometry):
-            raise ValueError("length of geometry and atomic_numbers does not match!")
+        if checks:
+            if len(atomic_numbers) != len(geometry):
+                raise ValueError("length of geometry and atomic_numbers does not match!")
 
-        if not all(isinstance(z, np.int8) for z in atomic_numbers) or atomic_numbers.size == 0:
-            raise ValueError("invalid atom list")
+            if not all(isinstance(z, np.int8) for z in atomic_numbers) or atomic_numbers.size == 0:
+                raise ValueError("invalid atom list")
 
-        if len(geometry) == 0:
-            raise ValueError("invalid geometry list")
+            if len(geometry) == 0:
+                raise ValueError("invalid geometry list")
 
-        try:
-            geometry = np.array(geometry)
-            geometry = geometry.astype(float)
-        except:
-            raise TypeError("geometry cannot be cast to ``np.ndarray`` of floats!")
-
-        if not all(all(isinstance(y, float) for y in x) for x in geometry):
-            raise TypeError("each element of self.geometry must be a 3-tuple")
-
-        if name and not isinstance(name, str):
-            raise TypeError("name must be a string!")
-
-        for atom in atomic_numbers:
             try:
-                get_symbol(atom)
-            except ValueError:
-                raise ValueError(f"unknwon atomic number {atom}")
-
-        if not isinstance(charge, int):
-            try:
-                charge = int(charge)
+                geometry = np.array(geometry)
+                geometry = geometry.astype(float)
             except:
-                raise TypeError("charge must be integer or castable to integer!")
+                raise TypeError("geometry cannot be cast to ``np.ndarray`` of floats!")
 
-        if not isinstance(multiplicity, int):
-            try:
-                multiplicity = int(multiplicity)
-            except:
-                raise TypeError("multiplicity must be positive integer or castable to positive integer")
-        assert multiplicity > 0, "multiplicity must be positive"
+            if not all(all(isinstance(y, float) for y in x) for x in geometry):
+                raise TypeError("each element of self.geometry must be a 3-tuple")
+
+            if name and not isinstance(name, str):
+                raise TypeError("name must be a string!")
+
+            for atom in atomic_numbers:
+                try:
+                    get_symbol(atom)
+                except ValueError:
+                    raise ValueError(f"unknwon atomic number {atom}")
+
+            if not isinstance(charge, int):
+                try:
+                    charge = int(charge)
+                except:
+                    raise TypeError("charge must be integer or castable to integer!")
+
+            if not isinstance(multiplicity, int):
+                try:
+                    multiplicity = int(multiplicity)
+                except:
+                    raise TypeError("multiplicity must be positive integer or castable to positive integer")
+            assert multiplicity > 0, "multiplicity must be positive"
 
         self.atomic_numbers = atomic_numbers.view(OneIndexedArray)
         self.geometry = geometry.view(OneIndexedArray)
 
-        if bonds:
-            for bond in bonds:
-                if len(bond) != 2:
-                    raise ValueError("while 3-center bonding is possible, it's a no-go in cctk")
-                self._check_atom_number(bond[0])
-                self._check_atom_number(bond[1])
-
         self.name = name
         self.multiplicity = multiplicity
         self.charge = charge
 
-        self.bonds = nx.Graph()
-        self.bonds.add_nodes_from(range(1, len(atomic_numbers) + 1))
-        if bonds:
+        if isinstance(bonds, nx.Graph):
+            self.bonds = bonds
+        elif isinstance(bonds, (list,np.ndarray,nx.classes.reportviews.EdgeView)):
+            if checks:
+                for bond in bonds:
+                    assert isinstance(bond, (tuple,list,np.ndarray)), f"expected a 2-iterable for bond, got {type(bond)}"
+                    assert len(bond)==2, f"unexpected number of atoms in bond, expected 2, got {len(bond)}"
+                    self._check_atom_number(bond[0])
+                    self._check_atom_number(bond[1])
+            self.bonds = nx.Graph()
+            self.bonds.add_nodes_from(range(1, len(atomic_numbers) + 1))
             for bond in bonds:
                 self.add_bond(bond[0], bond[1])
+        elif bonds is None:
+            self.bonds = nx.Graph()
+            self.bonds.add_nodes_from(range(1, len(atomic_numbers)+1))
+        else:
+            raise ValueError(f"unexpected type for bonds: {type(bonds)}")
 
     def __str__(self):
         if self.name is not None:
@@ -106,7 +115,10 @@ class Molecule:
         else:
             return f"Molecule ({len(self.atomic_numbers)} atoms)"
 
-    def assign_connectivity(self, cutoff=0.5):
+    def __hash__(self):
+        return hash(id(self))
+
+    def assign_connectivity(self, cutoff=0.1):
         """
         Automatically recalculates bonds based on covalent radii. If two atoms are closer than the sum of their covalent radii + 0.5 Angstroms, then they are considered bonded.
 
@@ -117,6 +129,8 @@ class Molecule:
             self
         """
 
+        assert isinstance(cutoff, (float, int)), "need cutoff to be numeric!"
+
         for i in range(1, self.num_atoms() + 1):
             for j in range(i + 1, self.num_atoms() + 1):
                 distance = self.get_distance(i, j)
@@ -165,7 +179,8 @@ class Molecule:
 
                 # 0.5 A distance is used by RasMol and Chime (documentation available online) and works well, empirically
                 if distance < (r_i + r_j - min_buffer):
-                    raise ValueError(f"atoms {i} and {j} are too close - distance {distance} A!")
+#                    raise ValueError(f"atoms {i} and {j} are too close - distance {distance} A!")
+                    return False
 
         return True
 
@@ -276,6 +291,8 @@ class Molecule:
         self._check_atom_number(atom1)
         self._check_atom_number(atom2)
 
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+
         bond_order = self.get_bond_order(atom1, atom2)
         if self.bonds.has_edge(atom1, atom2):
             self.bonds.remove_edge(atom1, atom2)
@@ -287,13 +304,14 @@ class Molecule:
             for fragment in fragments:
                 if atom1 in fragment:
                     if atom2 in fragment:
+                        self.add_bond(atom1, atom2, bond_order) # not adding back this bond causes some pretty pernicious errors
                         raise ValueError(f"Atom {atom1} and atom {atom2} are in a ring or otherwise connected!")
                     else:
                         fragment1 = fragment
                 if atom2 in fragment:
                     fragment2 = fragment
 
-            self.bonds.add_edge(atom1, atom2, weight=bond_order)
+            self.add_bond(atom1, atom2, bond_order)
             return list(fragment1), list(fragment2)
         else:
             raise ValueError(f"No bond between atom {atom1} and atom {atom2}!")
@@ -310,6 +328,7 @@ class Molecule:
         """
 
         self._check_atom_number(atom)
+
         fragments = nx.connected_components(self.bonds)
 
         for fragment in fragments:
@@ -317,7 +336,7 @@ class Molecule:
                 return list(fragment)
                 break
 
-    def set_distance(self, atom1, atom2, distance, move="group"):
+    def set_distance(self, atom1=None, atom2=None, distance=None, move="group", atoms=None):
         """
         Adjusts the ``atom1`` -- ``atom2`` bond length to be a fixed distance by moving atom2.
 
@@ -330,11 +349,20 @@ class Molecule:
             atom2 (int): the number of the second atom
             distance (float): distance in Angstroms of the final bond
             move (str): determines how fragment moving is handled
+            atoms (list): 2-element list of atom numbers
 
         Returns:
             the Molecule object
         """
 
+        if (atom1 is None) and (atom2 is None):
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 2, "need 2 atom numbers to set distance"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+
+        assert isinstance(distance, (float, int)), "need distance to set distance"
+
         self._check_atom_number(atom1)
         self._check_atom_number(atom2)
 
@@ -378,7 +406,7 @@ class Molecule:
 
         return self
 
-    def set_angle(self, atom1, atom2, atom3, angle, move="group"):
+    def set_angle(self, atom1=None, atom2=None, atom3=None, angle=None, move="group", atoms=None):
         """
         Adjusts the ``atom1`` -- ``atom2`` -- ``atom3`` bond angle to be a fixed value by moving ``atom3``.
 
@@ -392,11 +420,21 @@ class Molecule:
             atom3 (int): the number of the third atom
             angle (float): final value in degrees of the ``atom1`` -- ``atom2`` -- ``atom3`` angle
             move (str): determines how fragment moving is handled
+            atoms (list): 3-element list of atom numbers
 
         Returns:
             the Molecule object
         """
 
+        if (atom1 is None) and (atom2 is None) and (atom3 is None) :
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 3, "need 3 atom numbers to set angle"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+            atom3 = atoms[2]
+
+        assert isinstance(angle, (float, int)), "need angle to set angle"
+
         self._check_atom_number(atom1)
         self._check_atom_number(atom2)
         self._check_atom_number(atom3)
@@ -470,7 +508,7 @@ class Molecule:
 
         return self
 
-    def set_dihedral(self, atom1, atom2, atom3, atom4, dihedral, move="group34", check_result=True):
+    def set_dihedral(self, atom1=None, atom2=None, atom3=None, atom4=None, dihedral=None, move="group34", check_result=True, atoms=None):
         """
         Adjusts the ``atom1`` -- ``atom2`` -- ``atom3`` -- ``atom4`` dihedral angle to be a fixed value by moving atom 4.
 
@@ -488,16 +526,37 @@ class Molecule:
             dihedral (float): final value in degrees of the ``atom1`` -- ``atom2`` -- ``atom3`` -- ``atom4`` angle
             move (str): determines how fragment moving is handled
             check_result (Bool): whether the final answer should be checked for correctness
+            atoms (list): 4-element list of atomic numbers
 
         Returns:
             the Molecule object
         """
 
+        if (atom1 is None) and (atom2 is None) and (atom3 is None) and (atom4 is None):
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 4, "need 4 atom numbers to set dihedral"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+            atom3 = atoms[2]
+            atom4 = atoms[3]
+
+        assert isinstance(dihedral, (float, int)), "need angle to set dihedral angle"
+
+        # check atom numbers
         self._check_atom_number(atom1)
         self._check_atom_number(atom2)
         self._check_atom_number(atom3)
         self._check_atom_number(atom4)
 
+        # check there is bond connectivity information
+        assert len(self.bonds) > 0, "no bond connectivity information"
+
+        # check for collinearity
+        angle = self.get_angle(atom1, atom2, atom3, check=False)
+        assert 0.0001 < angle < 179.9999, f"1/2/3 atoms {atom1}-{atom2}-{atom3} are collinear (angle={angle:.8f})"
+        angle = self.get_angle(atom2, atom3, atom4, check=False)
+        assert 0.0001 < angle < 179.9999, f"2/3/4 atoms {atom2}-{atom3}-{atom4} are collinear (angle={angle:.8f})"
+
         for x in [atom1, atom2, atom3, atom4]:
             for y in [atom1, atom2, atom3, atom4]:
                 if x <= y:
@@ -560,7 +619,7 @@ class Molecule:
         delta = (dihedral - current_dihedral) % 360
 
         if np.abs(delta) < 0.001:
-            return
+            return self
 
         #### now the real work begins...
         #### move everything to place atom2 at the origin
@@ -730,14 +789,14 @@ class Molecule:
             the Molecule object
         """
 
-        if (not isinstance(coordinates, list)) or (len(coordinates) != 3):
+        if (not isinstance(coordinates, (list, np.ndarray)) or (len(coordinates) != 3)):
             raise TypeError("coordinates must be list with three elements")
 
         if not isinstance(symbol, str):
             raise TypeError(f"symbol {symbol} must be a string!")
 
         number = get_number(symbol)
-        self.atomic_numbers = np.append(self.atomic_numbers, [number]).view(OneIndexedArray)
+        self.atomic_numbers = np.append(self.atomic_numbers, [number]).astype(np.int8).view(OneIndexedArray)
         self.geometry = np.append(self.geometry, [coordinates], axis=0).view(OneIndexedArray)
         self.bonds.add_node(self.num_atoms())
 
@@ -833,7 +892,7 @@ class Molecule:
         else:
             return self.geometry[atom].view(np.ndarray)
 
-    def get_distance(self, atom1, atom2, check=True, _dist=compute_distance_between):
+    def get_distance(self, atom1=None, atom2=None, check=True, _dist=compute_distance_between, atoms=None):
         """
         Wrapper to compute distance between two atoms.
 
@@ -844,10 +903,17 @@ class Molecule:
             atom2 (int): number of the second atom
             check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
             _dist (function): function usd to compute distance
+            atoms (list): list of atomic numbers
 
         Returns:
             the distance, in Angstroms
         """
+        if (atom1 is None) and (atom2 is None):
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 2, "need 2 atom numbers to get distance"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+
         if check:
             try:
                 atom1 = int(atom1)
@@ -884,7 +950,7 @@ class Molecule:
 
         return np.sum(np.square(self.get_vector(atom1, atom2, check=False)))
 
-    def get_angle(self, atom1, atom2, atom3, check=True, _angle=compute_angle_between):
+    def get_angle(self, atom1=None, atom2=None, atom3=None, check=True, _angle=compute_angle_between, atoms=None):
         """
         Wrapper to compute angle between three atoms.
 
@@ -896,10 +962,18 @@ class Molecule:
             atom3 (int): number of the third atom
             check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
             _angle (function): function usd to compute angle
+            atoms (list): list of atom numbers
 
         Returns:
             the angle, in degrees
         """
+        if (atom1 is None) and (atom2 is None) and (atom3 is None):
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 3, "need 3 atom numbers to get angle"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+            atom3 = atoms[2]
+
         if check:
             try:
                 atom1 = int(atom1)
@@ -918,7 +992,7 @@ class Molecule:
 
         return _angle(v1 - v2, v3 - v2)
 
-    def get_dihedral(self, atom1, atom2, atom3, atom4, check=True, _dihedral=compute_dihedral_between):
+    def get_dihedral(self, atom1=None, atom2=None, atom3=None, atom4=None, check=True, _dihedral=compute_dihedral_between, atoms=None):
         """
         Wrapper to compute dihedral angle between four atoms.
 
@@ -931,10 +1005,19 @@ class Molecule:
             atom4 (int): number of the fourth atom
             check (Bool): whether to validate input data (can be overridden to prevent slow double-checking)
             _dihedral (function): function used to compute dihedral
+            atoms (list): list of atom numbers
 
         Returns:
             the dihedral angle, in degrees
         """
+        if (atom1 is None) and (atom2 is None) and (atom3 is None) and (atom4 is None):
+            assert isinstance(atoms, (list, np.ndarray)), "atom numbers need to come from fields or list!"
+            assert len(atoms) == 4, "need 4 atom numbers to get dihedral angle"
+            atom1 = atoms[0]
+            atom2 = atoms[1]
+            atom3 = atoms[2]
+            atom4 = atoms[3]
+
         if check:
             try:
                 atom1 = int(atom1)
@@ -1044,7 +1127,7 @@ class Molecule:
 
         return math.sqrt(distance) / self.num_atoms()
 
-    def optimize_dihedral(self, atom1, atom2, atom3, atom4):
+    def optimize_dihedral(self, atom1, atom2, atom3, atom4, step=10):
         """
         Minimizes the value of ``self.rms_distance_between_atoms`` for the given dihedral, in one-degree increments.
         A cheap alternative to geometry optimization using *ab initio* methods or density functional theory.
@@ -1054,6 +1137,7 @@ class Molecule:
             atom2 (int): atom number of the second atom in the dihedral
             atom3 (int): atom number of the third atom in the dihedral
             atom4 (int): atom number of the fourth atom in the dihedral
+            step (float): explore angles from 0 to 360 with this stepsize in degrees
 
         Returns:
             the final value of the angle
@@ -1066,7 +1150,7 @@ class Molecule:
         best_angle = 0
         best_dist = 0
 
-        for angle in range(0, 360):
+        for angle in range(0, 360, step):
             self.set_dihedral(atom1, atom2, atom3, atom4, angle)
             if self.rms_distance_between_atoms() > best_dist:
                 best_dist = self.rms_distance_between_atoms()
@@ -1153,3 +1237,337 @@ class Molecule:
         """
         hull = scipy.spatial.ConvexHull(self.geometry.view(np.ndarray))
         return hull.volume
+
+    def swap_atom_numbers(self, atom1, atom2):
+        """
+        Interchanges the numbers of ``atom1`` and ``atom2``.
+
+        Args:
+            atom1 (int): number of 1st atom
+            atom2 (int): number of 2nd atom
+
+        Returns
+            new ``Molecule`` object (does not modify in-place)
+        """
+        self._check_atom_number(atom1)
+        self._check_atom_number(atom2)
+        mol = copy.deepcopy(self)
+
+        z1 = mol.atomic_numbers[atom1]
+        z2 = mol.atomic_numbers[atom2]
+        g1 = copy.deepcopy(mol.geometry[atom1])
+        g2 = copy.deepcopy(mol.geometry[atom2])
+
+        mol.atomic_numbers[atom2] = z1
+        mol.atomic_numbers[atom1] = z2
+        mol.geometry[atom2] = g1
+        mol.geometry[atom1] = g2
+
+        mapping = {atom2: atom1, atom1: atom2}
+        mol.bonds = nx.relabel_nodes(mol.bonds, mapping, copy=True)
+        return mol
+
+    def epimerize(self, center_atom, substituent1, substituent2):
+        """
+        Epimerizes ``center_atom`` by exchanging the groups corresponding to ``substituent1`` and ``substituent2``.
+        Both substituents must be bonded to the center atom!
+
+        Args:
+            center_atom (int): number of middle atom
+            substituent1 (int): number of 1st atom
+            substituent1 (int): number of 2nd atom
+
+        Returns
+            new ``Molecule`` object (does not modify in-place)
+        """
+
+        self._check_atom_number(center_atom)
+        self._check_atom_number(substituent1)
+        self._check_atom_number(substituent2)
+
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+
+        adj = self.get_adjacent_atoms(center_atom)
+        assert len(adj) == 4, "center atom must be making 4 bonds!"
+        assert substituent1 in adj, "1st substituent is not bonded to center atom!"
+        assert substituent2 in adj, "2nd substituent is not bonded to center atom!"
+
+        #### remove both substituents
+        mol, group1, mmap1, gmap1  = cctk.Group.remove_group_from_molecule(self, center_atom, substituent1, return_mapping=True)
+        mol, group2, mmap2, gmap2  = cctk.Group.remove_group_from_molecule(mol, mmap1[center_atom], mmap1[substituent2], return_mapping=True)
+
+        h1 = mol.num_atoms() - 1
+        h2 = mol.num_atoms()
+
+        #### add them back in the opposite fashion
+        mol, mmap3, gmap3 =  cctk.Group.add_group_to_molecule(mol, group2, h1, return_mapping=True)
+        mol = cctk.Group.add_group_to_molecule(mol, group1, mmap3[h2])
+
+        #### relabel new graph to match original molecule
+        which = self._get_stereogenic_centers()
+        which.remove(center_atom)
+        return mol.renumber_to_match(self, check_chirality=which)
+
+    def renumber_to_match(self, model, check_chirality="all"):
+        """
+        Renumbers atoms to match ``model`` (must have isomorphic bond graph). Returns a copy of ``self`` with renumbered atoms.
+
+        Args:
+            model (cctk.Molecule): isomorphic molecule to renumber by
+            check_chirality (list of atomic numbers): atomic numbers to check, to prevent inversion due to graph isomorphism.
+                Alternatively ``None`` will prevent any checking and "all" will use ``self._get_exchangable_centers()``.
+
+        Returns:
+            new ``Molecule`` object
+        """
+
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+
+        #### use networkx to generate mapping
+        #### you need the node matcher to distinguish between e.g. H, F, Cl
+        self._add_atomic_numbers_to_nodes()
+        model._add_atomic_numbers_to_nodes()
+        nm = nx.algorithms.isomorphism.categorical_node_match("atomic_number", 0)
+
+        match = nx.algorithms.isomorphism.GraphMatcher(model.bonds, self.bonds, node_match=nm)
+        assert match.is_isomorphic(), "can't renumber non-isomorphic graphs!"
+        new_ordering = [match.mapping[x] for x in range(1, self.num_atoms() + 1)]
+        inv_mapping = {v:k  for k,v in match.mapping.items()} # bit kludgy but works
+
+        #### create renumbered molecule
+        mol = copy.deepcopy(self)
+        mol.atomic_numbers = self.atomic_numbers[new_ordering]
+        mol.geometry = self.geometry[new_ordering]
+        mol.bonds = nx.relabel_nodes(self.bonds, mapping=inv_mapping, copy=True)
+
+        if check_chirality == "all":
+            check_chirality = mol._get_exchangeable_centers()
+
+        #### diastereotopic protons get scrambled by the above code so we gotta go through and fix all of them
+        #### this happens because networkx doesn't store chirality - a known limitation of graph molecular encoding!
+        if isinstance(check_chirality, list):
+            #### find all the differences and exchange them
+            model_report = model.get_chirality_report(check_chirality)
+            mol_report = mol.get_chirality_report(check_chirality)
+
+            #### generate all meso ring permutations
+            candidates = mol.flip_meso_rings(atoms=check_chirality)
+
+            #### for each, try flipping configuration of all centers
+            for candidate in candidates:
+                for center in check_chirality:
+                    if model_report[center] != candidate.get_chirality_report(check_chirality)[center]:
+                        try:
+                            candidate = candidate.exchange_identical_substituents(center)
+                        except ValueError as e:
+                            break
+
+                #### check that we actually fixed all the problems
+                mol_report = candidate.get_chirality_report(check_chirality)
+                all_good = True
+                for center in check_chirality:
+                    if mol_report[center] != model_report[center]:
+                        all_good = False
+                        break
+                #### if we did, then return
+                if all_good:
+                    return candidate
+
+        raise ValueError("can't get a proper renumbering: are you *sure* these two molecules can have the same chirality?")
+
+    def _get_stereogenic_centers(self):
+        """
+        Returns every atom making 4 or more bonds. A bit misleading, since diastereotopic protons/meso protons are also counted.
+        """
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+        num_neighbors = np.array([len(list(self.bonds[x])) for x in range(1, self.num_atoms() + 1)])
+        return [int(x) for x in list(np.ravel(np.argwhere(num_neighbors >= 4)) + 1)] # love me some off-by-one indexing errors
+
+    def _get_exchangeable_centers(self):
+        """
+        Returns all atoms making 4 or more bonds that have two isomorphic substituents, i.e. where renumbering could be broken.
+        """
+        centers = self._get_stereogenic_centers()
+        exchangeable_centers = []
+        for center in centers:
+            try:
+                mol = self.exchange_identical_substituents(center)
+                exchangeable_centers.append(center)
+                continue
+            except:
+                pass
+
+            mols = self.flip_meso_rings(atoms=[center])
+            if len(mols) > 1:
+                exchangeable_centers.append(center)
+
+        return exchangeable_centers
+
+    def get_chirality_report(self, centers=None):
+        """
+        Computes chirality at stereogenic centers.
+
+        Args:
+            centers (list): atomic numbers to check. defaults to all centers with 4+ substituents.
+
+        Returns:
+            dict with centers as keys and ±1 as values
+        """
+        if centers is None:
+            centers = self._get_stereogenic_centers()
+        assert isinstance(centers, list)
+
+        results = {}
+        for center in centers:
+            neighbors = list(self.bonds[center])
+            neighbors.sort()
+            assert len(neighbors) >= 4, f"atom {center} has fewer than 4 neighbors ({neighbors})!"
+            results[center] = compute_chirality(*[self.get_vector(n, center) for n in neighbors])
+
+        return results
+
+    def exchange_identical_substituents(self, center, self_permutations=None):
+        """
+        Replace homotopic/enantiotopic/diastereotopic substituents about a single atom.
+
+        If a list of permuted ``Molecule`` objects is passed (as ``self_permutations``), then this code will apply this to each member and return a list.
+
+        Args:
+            center (integer): atomic number of atom to swap substituents around
+            self_permutations (list of Molecules): optional list of starting ``Molecule`` objects
+
+        Returns:
+            ``Molecule`` object (or list if ``self_permutations`` is not ``None``)
+        """
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+        self._add_atomic_numbers_to_nodes()
+        neighbors = list(self.bonds[center])
+
+        returns = [copy.deepcopy(self)]
+        if self_permutations is not None:
+            returns = self_permutations
+
+        for i in range(len(neighbors)):
+            for j in range(i+1, len(neighbors)):
+                try:
+                    _, frag1 = self._get_bond_fragments(center, neighbors[i])
+                    _, frag2 = self._get_bond_fragments(center, neighbors[j])
+
+                    graph1 = self.bonds.subgraph(frag1)
+                    graph2 = self.bonds.subgraph(frag2)
+
+                    nm = nx.algorithms.isomorphism.categorical_node_match("atomic_number", 0)
+                    match = nx.algorithms.isomorphism.GraphMatcher(graph1, graph2, node_match=nm)
+
+                    if match.is_isomorphic():
+                        new_returns = []
+                        for mol in returns:
+                            new_mol = copy.deepcopy(mol)
+                            for k,v in match.mapping.items():
+                                new_mol = new_mol.swap_atom_numbers(k, v)
+                            if self_permutations is None:
+                                return new_mol
+
+                        returns.append(new_mol)
+
+                except ValueError as e:
+                    pass # probably indicates a cycle
+
+        if self_permutations is None:
+            raise ValueError("could not find substituents to switch")
+        else:
+            return returns
+
+    def _add_atomic_numbers_to_nodes(self):
+        """
+        Add the atomic numbers to each node attribute, to allow for distinguishment of F and H during graph renumbering.
+        """
+        nx.set_node_attributes(self.bonds, {z: {"atomic_number": self.atomic_numbers[z]} for z in range(1, self.num_atoms() +  1)})
+
+    def is_atom_in_ring(self, atom):
+        assert self.bonds.number_of_edges() > 0, "need a bond graph to perform this operation -- try calling self.assign_connectivity()!"
+        cycles = nx.cycle_basis(self.bonds, root=atom)
+        for cycle in cycles:
+            if atom in cycle:
+                return True
+        return False
+
+    def flip_meso_rings(self, atoms):
+        """
+        Returns a list of permuted molecules with various ``meso`` rings renumbered.
+
+        Args:
+            atoms (list): atomic numbers of potential atoms to consider
+
+        Returns:
+            list of ``Molecule`` objects
+        """
+        #### get all rings in graph
+        returns = [copy.deepcopy(self)]
+        for center in atoms:
+            cycles = nx.cycle_basis(self.bonds, root=center)
+            for cycle in cycles:
+                #### get the correct ring
+                if center not in cycle:
+                    continue
+
+                #### reorder to put ``center`` first
+                idx = cycle.index(center)
+                while cycle[0] != center:
+                    # why yes, this /is/ a O(n) solution for reordering a list. why do you ask?
+                    cycle = cycle[1:] + cycle[0:1]
+                assert cycle[0] == center, "graph reorder failed"
+
+                #### create fragments
+                frag1 = [cycle.pop(1)]
+                frag2 = [cycle.pop(-1)]
+                while len(cycle) > 2:
+                    frag1.append(cycle.pop(1))
+                    frag2.append(cycle.pop(-1))
+
+                #### cut fragment bonds, depending on if we have even- or odd-numbered ring
+                new_returns = []
+                for mol in returns:
+                    mol = copy.deepcopy(mol)
+                    mol.remove_bond(frag1[0], cycle[0])
+                    mol.remove_bond(frag2[0], cycle[0])
+                    if len(cycle) == 1:
+                        mol.remove_bond(frag1[-1], frag2[-1])
+                    elif len(cycle) == 2:
+                        mol.remove_bond(frag1[-1], cycle[-1])
+                        mol.remove_bond(frag2[-1], cycle[-1])
+
+                    #### generate graphs
+                    graph1 = None
+                    graph2 = None
+                    fragments = nx.connected_components(mol.bonds)
+                    for fragment in fragments:
+                        if frag1[0] in fragment:
+                            graph1 = mol.bonds.subgraph(fragment)
+                        if frag2[0] in fragment:
+                            graph2 = mol.bonds.subgraph(fragment)
+
+                    assert isinstance(graph1, nx.Graph), "can't find graph 1"
+                    assert isinstance(graph2, nx.Graph), "can't find graph 1"
+
+                    #### do our two ring-halves match?? if so, we swap them
+                    nm = nx.algorithms.isomorphism.categorical_node_match("atomic_number", 0)
+                    match = nx.algorithms.isomorphism.GraphMatcher(graph1, graph2, node_match=nm)
+
+                    if match.is_isomorphic():
+                        for k,v in match.mapping.items():
+                            mol = mol.swap_atom_numbers(k, v)
+
+                        #### redo all the bonds we ablated
+                        if len(cycle) == 1:
+                            mol.add_bond(frag1[-1], frag2[-1], self.get_bond_order(frag1[-1], frag2[-1]))
+                        elif len(cycle) == 2:
+                            mol.add_bond(frag1[-1], cycle[-1], self.get_bond_order(frag1[-1], cycle[-1]))
+                            mol.add_bond(frag2[-1], cycle[-1], self.get_bond_order(frag2[-1], cycle[-1]))
+                        mol.add_bond(frag1[0], cycle[0], self.get_bond_order(frag1[0], cycle[0]))
+                        mol.add_bond(frag2[0], cycle[0], self.get_bond_order(frag2[0], cycle[0]))
+
+                        new_returns.append(mol)
+                returns = returns + new_returns
+        return returns
+
