@@ -2,13 +2,12 @@ import numpy as np
 import re
 
 from cctk.helper_functions import get_symbol
-from cctk import OneIndexedArray
+from cctk import OneIndexedArray, LazyLineObject
 
 """
 Functions to help with parsing Gaussian files
 """
 
-
 def read_geometries_and_energies(lines):
     """
     A large and unwieldy method - reads geometries, symbol lists, and energies from the file.
@@ -33,6 +32,10 @@ def read_geometries_and_energies(lines):
     this_energy = None
 
     i = 0
+
+    #### we'll read this into memory to speed this up
+    lines = list(lines)
+
     in_geometry_block = False
     while i < len(lines):
         # read the current line
@@ -119,52 +122,6 @@ def read_geometries_and_energies(lines):
         (geometry, symbol_list) = extract_initial_geometry(lines)
         return [geometry], symbol_list, [], []
 
-
-def search_for_block(lines, start, end, count=1, join=""):
-    """
-    Search through a file (lines) and locate a block starting with "start" (inclusive) and ending with "end" (exclusive).
-
-    Args:
-        lines (list): a list of the lines in the file
-        start (str): a pattern that matches the start of the block (can contain special characters)
-        end (str): a pattern that matches the end of the block (can contain special characters)
-        count (int): how many matches to search for
-        join (str): spacer between lines
-
-    Returns:
-        a single match (str) if count == 1 or a list of matches (str) if count > 1.
-    """
-    assert isinstance(count, int), "count needs to be an integer"
-    assert isinstance(join, str), "join needs to be a string"
-
-    current_match = ""
-    match = [None] * count
-
-    start_pattern = re.compile(start)
-    end_pattern = re.compile(end)
-
-    index = 0
-    for line in lines:
-        if current_match:
-            if end_pattern.search(line):
-                match[index] = current_match
-                current_match = None
-                index += 1
-
-                if index == count:
-                    break
-            else:
-                current_match = current_match + join + line.lstrip()
-        else:
-            if start_pattern.search(line):
-                current_match = line.lstrip()
-
-    if count == 1:
-        return match[0]
-    else:
-        return match
-
-
 def read_bonds(lines):
     """
     Reads bonding information from the output file.
@@ -181,6 +138,10 @@ def read_bonds(lines):
 
     i = 0
     in_bonding_section = False
+
+    #### we'll read this into memory to speed this up
+    lines = list(lines)
+
     while i < len(lines):
         # read the current line
         line = lines[i].strip()
@@ -217,54 +178,6 @@ def read_bonds(lines):
     else:
         return None
 
-
-def find_parameter(lines, parameter, expected_length, which_field, split_on=None):
-    """
-    Helper method to search through the output file and find key forces and displacements.
-
-    Args:
-        lines (list): list of lines in file
-        parameter (string): test to search for
-        expected_length (int): how many fields there should be
-        which_field (int): which field the parameter is (zero-indexed)
-        split_on (str): additional non-space field to split
-    Returns:
-        a list of all the extracted values
-    """
-
-    if (not isinstance(expected_length, int)) or (not isinstance(which_field, int)):
-        raise TypeError("expected_length and which_field must be type int!")
-
-    if which_field >= expected_length:
-        raise ValueError("can't expect a field after the last field!")
-
-    matches = []
-    pattern = False
-
-    try:
-        pattern = re.compile(parameter)
-    except:
-        raise ValueError("pattern {pattern} cannot be compiled as a regex; try again!")
-
-    if pattern:
-        for line in lines:
-            if pattern.search(line):
-                fields = re.split(" +", line)
-                if split_on:
-                    fields2 = []
-                    for field in fields:
-                        fields2 = fields2 + field.split(split_on)
-                    fields = fields2
-                fields = list(filter(None, fields))
-
-                if len(fields) == expected_length:
-                    try:
-                        matches.append(float(fields[which_field]))
-                    except:
-                        pass
-        return matches
-
-
 def extract_initial_geometry(lines):
     """
     Helper method to search through the output file and find the initial geometry/symbol list, in cases where no SCF iterations finished.
@@ -276,7 +189,7 @@ def extract_initial_geometry(lines):
         initial geometry (list)
         symbol list (list)
     """
-    initial_geom_block = search_for_block(lines, "Symbolic Z-matrix:", "^ $", join="\n")
+    initial_geom_block = lines.search_for_block("Symbolic Z-matrix:", "^ $", join="\n")
     atom_lines = initial_geom_block.split("\n")[2:]
     geometry = [None] * len(atom_lines)
     symbols = [None] * len(atom_lines)
@@ -314,27 +227,26 @@ def read_nmr_shifts(lines, num_atoms):
     assert len(shieldings) == num_atoms, f"Expected {num_atoms} shieldings but found {len(shieldings)}!"
     return np.asarray(shieldings)
 
-def split_link1(lines):
+def split_link1(filename):
     """
-    Splits ``lines`` list into blocks by searching for "Entering Link 1".
+    Splits ``filename`` into blocks by searching for "Entering Link 1".
 
     Args:
-        lines (list): list of lines in file
+        filename (str): path to file
 
     Returns:
         list of list of lines by Link1 section; so a file with one Link1 specification would return [lines1, lines2]
     """
     link1_blocks = []
-    current_block = []
 
-    for line in lines:
-        if re.search("Entering Link 1", line):
-            link1_blocks.append(current_block)
-            current_block = []
-        else:
-            current_block.append(line)
+    start_block = 0
+    with open(filename, "r") as lines:
+        for idx, line in enumerate(lines):
+            if re.search("Entering Link 1", line):
+                link1_blocks.append(LazyLineObject(file=filename, start=start_block, end=idx))
+                start_block = idx
+    link1_blocks.append(LazyLineObject(file=filename, start=start_block, end=idx))
 
-    link1_blocks.append(current_block)
     return link1_blocks[1:] #### the first block is just a few lines
 
 def extract_link0(lines):
@@ -348,12 +260,17 @@ def extract_link0(lines):
         dictionary of link 0 commands
     """
     output = {}
+
+    #### read first 200 lines of file to memory
+    lines = [x for _, x in zip(range(200), lines)]
+
     #### find where the header starts
     for idx, line in enumerate(lines):
         if re.match(" #p", line):
             break
     end_idx = idx - 1
 
+
     ##### then go back until the big header that says "GAUSSIAN 16"
     while not re.match(" \*\*\*\*\*\*\*\*", lines[idx]):
         idx += -1
@@ -380,7 +297,7 @@ def read_forces(lines):
         (n x 3) ``cctk.OneIndexedArray`` of forces
     """
     forces = []
-    force_block = search_for_block(lines, "Forces \(Hartrees/Bohr\)", "Cartesian Forces", join="\n")
+    force_block = lines.search_for_block("Forces \(Hartrees/Bohr\)", "Cartesian Forces", join="\n")
     for line in force_block.split("\n")[2:]:
         fields = re.split(" +", line)
         fields = list(filter(None, fields))
@@ -401,13 +318,14 @@ def read_mulliken_charges(lines):
         ``cctk.OneIndexedArray`` of charges
     """
     charges = []
-    charge_block = search_for_block(lines, " Mulliken charges:", " Sum of Mulliken charges", join="\n")
-    for line in charge_block.split("\n")[2:]:
-        fields = re.split(" +", line)
-        fields = list(filter(None, fields))
+    charge_block = lines.search_for_block(" Mulliken charges:", " Sum of Mulliken charges", join="\n")
+    if charge_block is not None:
+        for line in charge_block.split("\n")[2:]:
+            fields = re.split(" +", line)
+            fields = list(filter(None, fields))
 
-        if len(fields) == 3:
-            charges.append(float(fields[2]))
+            if len(fields) == 3:
+                charges.append(float(fields[2]))
 
     return OneIndexedArray(charges)
 
@@ -424,7 +342,7 @@ def read_hirshfeld_charges(lines):
     """
     charges = []
     spins = []
-    charge_block = search_for_block(lines, "Hirshfeld charges, spin densities, dipoles, and CM5 charges", " Hirshfeld charges", join="\n")
+    charge_block = lines.search_for_block("Hirshfeld charges, spin densities, dipoles, and CM5 charges", " Hirshfeld charges", join="\n")
     for line in charge_block.split("\n")[2:]:
         fields = re.split(" +", line)
         fields = list(filter(None, fields))
@@ -445,10 +363,11 @@ def read_dipole_moment(lines):
     Returns:
         dipole moment (magnitude only)
     """
-    dipole_block = search_for_block(lines, "Dipole moment \(field-independent basis, Debye\)", "Quadrupole moment \(field-independent basis, Debye-Ang\):", join="\n")
+    dipole_block = lines.search_for_block(" Electronic spatial extent", " Quadrupole moment", join="\n")
     for line in dipole_block.split("\n")[1:]:
         fields = re.split(" +", line)
         fields = list(filter(None, fields))
 
         if len(fields) == 8:
             return float(fields[7])
+
