@@ -26,11 +26,13 @@ class TestNMR(unittest.TestCase):
         gaussian_file = cctk.GaussianFile.read_file("test/static/methane2.out")
         self.assertEqual(len(gaussian_file), 2)
         first_link = gaussian_file[0]
+        self.assertLess(abs(first_link.elapsed_time-6.1), 0.001)
         self.assertListEqual(first_link.job_types, [JobType.OPT, JobType.FREQ, JobType.SP])
         ensemble = first_link.ensemble
         energies = list(ensemble[:,"energy"])
         self.assertListEqual(energies, [-40.5169484082, -40.5183831835, -40.5183831835])
         second_link = gaussian_file[1]
+        self.assertLess(abs(second_link.elapsed_time-1.9), 0.001)
         ensemble = second_link.ensemble
         shieldings = list(ensemble[-1,"isotropic_shielding"])
         self.assertListEqual(list(shieldings), [192.9242, 31.8851, 31.8851, 31.8851, 31.8851])
@@ -42,10 +44,12 @@ class TestNMR(unittest.TestCase):
 
         # opt freq
         first_link = gaussian_file[0]
+        self.assertLess(abs(first_link.elapsed_time-11.0), 0.001)
         self.assertListEqual(first_link.job_types, [JobType.OPT, JobType.FREQ, JobType.SP])
 
         # NMR: ethane
         second_link = gaussian_file[1]
+        self.assertLess(abs(second_link.elapsed_time-2.1), 0.001)
         ensemble = second_link.ensemble
         molecule = ensemble[-1]
         #shifts = list(ensemble[molecule]["isotropic_shielding"])
@@ -57,6 +61,7 @@ class TestNMR(unittest.TestCase):
 
         # NMR: methane
         third_link = gaussian_file[2]
+        self.assertLess(abs(third_link.elapsed_time-1.8), 0.001)
         self.assertListEqual(third_link.job_types, [JobType.NMR, JobType.SP])
         ensemble = third_link.ensemble
         #molecule = ensemble[-1]
@@ -69,6 +74,7 @@ class TestNMR(unittest.TestCase):
     def test_nmr4(self):
         # tests code for scaling NMR shieldings on a more complicated molecule
         gaussian_file = cctk.GaussianFile.read_file("test/static/LSD_custom.out")
+        self.assertEqual(gaussian_file.successful_terminations,1)
         ensemble = gaussian_file.ensemble
         shieldings = ensemble[:,"isotropic_shielding"]
         scaled_shifts, shift_labels = cctk.helper_functions.scale_nmr_shifts(ensemble,
